@@ -11,7 +11,6 @@ use vars qw($VERSION);
 $VERSION = '1.00';
 
 my @types = (
-  undef,
   [ "BYTE",      "C1", 1],
   [ "ASCII",     "A1", 1],
   [ "SHORT",     "n1", 2],
@@ -171,6 +170,7 @@ my %makernotes = (
     "OLYMPUS OPTICAL CO.,LTD C2500L"  => [8, 'Olympus', \%olympus_tags],
     "OLYMPUS OPTICAL CO.,LTD C3030Z"  => [8, 'Olympus', \%olympus_tags],
     "OLYMPUS OPTICAL CO.,LTD C3040Z"  => [8, 'Olympus', \%olympus_tags],
+    "OLYMPUS OPTICAL CO.,LTD C4100Z,C4000Z" => [8, 'Olympus', \%olympus_tags],
     "OLYMPUS OPTICAL CO.,LTD E-10"    => [8, 'Olympus', \%olympus_tags],
     "FUJIFILM FinePix4900ZOOM"  => [-1, 'FinePix', \%fujifilm_tags],
     "FUJIFILM FinePix6900ZOOM"  => [-1, 'FinePix', \%fujifilm_tags],
@@ -212,6 +212,7 @@ my %exif_tags = (
     0x8769 => "ExifOffset",
     0x8773 => "InterColorProfile",
     0x8822 => { __TAG__ => "ExposureProgram",
+		0 => "unknown",
 		1 => "Manual",
 		2 => "Program",
 		3 => "Aperture priority",
@@ -220,6 +221,7 @@ my %exif_tags = (
 		6 => "Program action",
 		7 => "Portrait",
 		8 => "Landscape",
+		# 9 .. 255 reserved
 	      },
     0x8824 => "SpectralSensitivity",
     0x8825 => "GPSInfo",
@@ -239,24 +241,67 @@ my %exif_tags = (
     0x9205 => "MaxApertureValue",
     0x9206 => "SubjectDistance",
     0x9207 => { __TAG__ => "MeteringMode",
+		0 => "unknown",
 		1 => "Average",
-		2 => "Center weighted average",
+		2 => "CenterWeightedAverage",
 		3 => "Spot",
-		4 => "Multi-spot",
-		5 => "Multi-segment",
+		4 => "MultiSpot",
+		5 => "Pattern",
+		6 => "Partial",
+		# 7 .. 254 reserved in EXIF 1.2
+		255 => "other",
 	      },
     0x9208 => { __TAG__ => "LightSource",
-		0 => "Auto",
+		0 => "unknown",
 		1 => "Daylight",
 		2 => "Fluorescent",
 		3 => "Tungsten",
-		10 => "Flash",
+		4 => "Flash",
+		# 5 .. 8 reserveed in EXIF 2.2
+		9 => "Fine weather",
+		10 => "Cloudy weather",
+		11 => "Shade",
+		12 => "Daylight fluorescent (D 5700-7100K)",
+		13 => "Day white fluorescent (N 4600-5400K)",
+		14 => "Cool white fluorescent (W 3900-4500K)",
+		15 => "White fluorescent (WW 3200-3700K)",		
+		17 => "Standard light A",
+		18 => "Standard light B",
+		19 => "Standard light C",
+		20 => "D55",
+		21 => "D65",
+		22 => "D75",
+		23 => "D50",
+		24 => "ISO studio tungesten",		
+		# 25 .. 254 reserved in EXIF 2.2
+		255 => "other light source",
 	      },
     0x9209 => { __TAG__ => "Flash",
-		0 => "No",
-		1 => "Yes",
-	      },
+		0x0000 => "Flash did not fire",
+		0x0001 => "Flash fired",
+		0x0005 => "Strobe return light not detected",
+		0x0007 => "Strobe return light detected",
+		0x0009 => "Flash fired, compulsory flash mode",
+		0x000D => "Flash fired, compulsory flash mode, return light not detected",
+		0x000F => "Flash fired, compulsory flash mode, return light detected",
+		0x0010 => "Flash did not fire, compulsory flash mode",
+		0x0018 => "Flash did not fire, auto mode",
+		0x0019 => "Flash fired, auto mode",
+		0x001D => "Flash fired, auto mode, return light not detected",
+		0x001F => "Flash fired, auto mode, return light detected",
+		0x0020 => "No flash function",
+		0x0041 => "Flash fired, red-eye reduction mode",
+		0x0045 => "Flash fired, red-eye reduction mode, return light not detected",
+		0x0047 => "Flash fired, red-eye reduction mode, return light detected",
+		0x0049 => "Flash fired, compulsory flash mode, red-eye reduction mode",
+		0x004D => "Flash fired, compulsory flash mode, red-eye reduction mode, return light not detected",
+		0x004F => "Flash fired, compulsory flash mode, red-eye reduction mode, return light detected",
+		0x0059 => "Flash fired, auto mode, red-eye reduction mode",
+		0x005D => "Flash fired, auto mode, return light not detected, red-eye reduction mode",
+		0x005F => "Flash fired, auto mode, return light detected, red-eye reduction mode"
+		},
     0x920A => "FocalLength",
+    0x9214 => "SubjectArea",
     0x927C => "MakerNote",
     0x9286 => "UserComment",
     0x9290 => "SubSecTime",
@@ -266,6 +311,7 @@ my %exif_tags = (
     0xA001 => "ColorSpace",
     0xA002 => "ExifImageWidth",
     0xA003 => "ExifImageLength",
+    0xA004 => "RelatedAudioFile",
     0xA005 => {__TAG__ => "InteroperabilityOffset",
 	       __SUBIFD__ => \%exif_intr_tags,
 	      },
@@ -282,16 +328,81 @@ my %exif_tags = (
 	      },
     0xA214 => "SubjectLocation",              # 0x9214    -  -
     0xA215 => "ExposureIndex",                # 0x9215    -  -
-    0xA217 => "SensingMethod",                # 0x9217    -  -
+    0xA217 => {__TAG__ => "SensingMethod",
+		1 => "Not defined",
+		2 => "One-chip color area sensor",
+		3 => "Two-chip color area sensor",
+		4 => "Three-chip color area sensor",
+		5 => "Color sequential area sensor",
+		7 => "Trilinear sensor",
+		8 => "Color sequential linear sensor"
+		},
     0xA300 => {__TAG__ => "FileSource",
                DECODER => \&file_source_decoder,
               },
     0xA301 => {__TAG__ => "SceneType",
                DECODER => \&scene_type_decoder,
               },
+    0xA302 => "CFAPattern",
+    0xA401 => {__TAG__ => "CustomRendered",
+		0 => "Normal process",
+		1 => "Custom process"
+		},
+    0xA402 => {__TAG__ => "ExposureMode",
+		0 => "Auto exposure",
+		1 => "Manual exposure",
+		2 => "Auto bracket"
+		},
+    0xA403 => {__TAG__ => "WhiteBalance",
+		0 => "Auto white balance",
+		1 => "Manual white balance"
+		},
+    0xA404 => "DigitalZoomRatio",
+    0xA405 => "FocalLengthIn35mmFilm",
+    0xA406 => {__TAG__ => "SceneCaptureType",
+		0 => "Standard",
+		1 => "Landscape",
+		2 => "Portrait",
+		3 => "Night Scene"
+		},
+    0xA407 => {__TAG__ => "GainControl",
+		0 => "None",
+		1 => "Low gain up",
+		2 => "High gain up",
+		3 => "Low gain down",
+		4 => "High gain down"
+		},
+    0xA408 => {__TAG__ => "Contrast",
+		0 => "Normal",
+		1 => "Soft",
+		2 => "Hard"
+		},
+    0xA409 => {__TAG__ => "Saturation",
+		0 => "Normal",
+		1 => "Low saturation",
+		2 => "High saturation"
+		},
+    0xA40A => {__TAG__ => "Sharpness",
+		0 => "Normal",
+		1 => "Soft",
+		2 => "Hard"
+		},
+    0xA40B => "DeviceSettingDescription",
+    0xA40C => {__TAG__ => "SubjectDistanceRange",
+		0 => "Unknown",
+		1 => "Macro",
+		2 => "Close view",
+		3 => "Distant view"
+		},
+    0xA420 => "ImageUniqueID",
 );
 
 my %tiff_tags = (
+  254   => { __TAG__ => "NewSubfileType",
+	     1 => "ReducedResolution",
+	     2 => "SinglePage",
+	     4 => "TransparencyMask",
+	 },
   255   => { __TAG__ => "SubfileType",
 	     1 => "FullResolution",
 	     2 => "ReducedResolution",
@@ -312,7 +423,7 @@ my %tiff_tags = (
   262   => { __TAG__ => "PhotometricInterpretation",
 	     0 => "WhiteIsZero",
 	     1 => "BlackIsZero",
-             2 => "RGB",
+	     2 => "RGB",
 	     3 => "RGB Palette",
 	     4 => "Transparency Mask",
 	     5 => "CMYK",
@@ -324,10 +435,14 @@ my %tiff_tags = (
 	     2 => "OrderedDither",
 	     3 => "Randomized",
 	   },
+  266   => { __TAG__ => "FillOrder",
+	     1 => "LowInHigh",
+	     2 => "HighInLow",
+	   },
   270   => "ImageDescription",
   271   => "Make",
   272   => "Model",
-  273   => "StipOffset",
+  273   => "StripOffsets",
   274   => { __TAG__ => "Orientation",
 	     1 => "top_left",
 	     2 => "top_right",
@@ -349,11 +464,20 @@ my %tiff_tags = (
   296   => {__TAG__ => "ResolutionUnit",
 	    1 => "pixels", 2 => "dpi", 3 => "dpcm",
 	   },
+  297   => "PageNumber",
+  301   => "TransferFunction",
   305   => "Software",
   306   => "DateTime",
+  315   => "Artist",
+  318   => "WhitePoint",
+  319   => "PrimaryChromaticities",
+  320   => "ColorMap",
   513   => "JPEGInterchangeFormat",
   514   => "JPEGInterchangeFormatLngth",
+  529   => "YCbCrCoefficients",
+  530   => "YCbCrSubSampling",
   531   => "YCbCrPositioning",
+  532   => "ReferenceBlackWhite",
   33432 => "Copyright",
   34665 => { __TAG__ => "ExifOffset",
 	     __SUBIFD__ => \%exif_tags,
@@ -405,6 +529,7 @@ sub unpack
     if ($self->{little_endian}) {
 	$template =~ tr/nN/vV/;
     }
+    #print "UNPACK $template\n";
     CORE::unpack($template, $_[0]);
 }
 
@@ -427,6 +552,11 @@ sub tagname
     $tiff_tags{$_[1]} || sprintf "Tag-0x%04x",$_[1];
 }
 
+sub exif_tagname
+{
+    $tiff_tags{$_[1]} || $exif_tags{$_[1]} || sprintf "Tag-0x%04x",$_[1];
+}
+
 sub add_fields
 {
     my($self, $offset, $ifds, $tags, $voff_plus) = @_;
@@ -434,32 +564,54 @@ sub add_fields
     $tags ||= \%tiff_tags;
 
     for (${$self->{source}}) {  # alias as $_
+	last if $offset > length($_) - 2;  # bad offset
 	my $entries = $self->unpack("x$offset n", $_);
+	my $max_entries = int((length($_) - $offset - 2) / 12);
+	#print "ENTRIES $entries $max_entries\n";
+	if ($entries > $max_entries) {
+	    # Hmmm, something smells bad here...  parsing garbage
+	    $entries = $max_entries;
+	    last;
+	}
       FIELD:
 	for my $i (0 .. $entries-1) {
+	    my $entry_offset = 2 + $offset + $i*12;
 	    my($tag, $type, $count, $voff) =
-		$self->unpack("nnNN", substr($_, 2 + $offset + $i*12, 12));
-	    $voff += $voff_plus || 0;
-	    my $val;
-	    if (my $t = $types[$type]) {
-		$type = $t->[0];
-		my $tmpl = $t->[1];
-		my $vlen = $t->[2];
-		if ($count * $vlen <= 4) {
-		    $voff = 2 + $offset + $i*12 + 8;
-		}
-		$tmpl =~ s/(\d+)$/$count*$1/e;
+		$self->unpack("nnNN", substr($_, $entry_offset, 12));
+	    #print "TAG $tag $type $count $voff\n";
 
-		my @v = $self->unpack("x$voff$tmpl", $_);
-
-		if ($type =~ /^S(SHORT|LONG|RATIONAL)/) {
-		    my $max = 2 ** ($1 eq "SHORT" ? 15 : 31);
-		    $v[0] -= ($max * 2) if $v[0] >= $max;
-		}
-
-		$val = (@v > 1) ? \@v : $v[0];
-		bless $val, "Image::TIFF::Rational" if $type =~ /^S?RATIONAL$/;
+	    if ($type == 0 || $type > @types) {
+		# unknown type code might indicate that we are parsing garbage
+		next;
 	    }
+
+	    # extract type information
+	    my($tmpl, $vlen);
+	    ($type, $tmpl, $vlen) = @{$types[$type-1]};
+	    die "Assert" unless $type;
+
+	    if ($count * $vlen <= 4) {
+		$voff = $entry_offset + 8;
+	    }
+	    elsif ($voff + $count * $vlen > length($_)) {
+		# offset points outside of string, corrupt entry ignore
+		next;
+	    }
+	    else {
+		$voff += $voff_plus || 0;
+	    }
+	    $tmpl =~ s/(\d+)$/$count*$1/e;
+
+	    my @v = $self->unpack("x$voff $tmpl", $_);
+
+	    if ($type =~ /^S(SHORT|LONG|RATIONAL)/) {
+		my $max = 2 ** ($1 eq "SHORT" ? 15 : 31);
+		$v[0] -= ($max * 2) if $v[0] >= $max;
+	    }
+
+	    my $val = (@v > 1) ? \@v : $v[0];
+	    bless $val, "Image::TIFF::Rational" if $type =~ /^S?RATIONAL$/;
+
 	    $tag = $tags->{$tag} || $self->tagname($tag);
 
 	    if ($tag eq 'MakerNote' && exists $makernotes{$self->{Make}.' '.$self->{Model}}) {
@@ -482,12 +634,14 @@ sub add_fields
 	    if (ref($tag)) {
 		die "Assert" unless ref($tag) eq "HASH";
 		if (my $sub = $tag->{__SUBIFD__}) {
+		    next if $val < 0 || $val > length($_);
+		    #print "SUBIFD $tag->{__TAG__} $val ", length($_), "\n";
 		    $self->add_fields($val, $ifds, $sub);
 		    next FIELD;
 		}
 		#hack for UNDEFINED values, they all have different
 		#meanings depending on tag
-		$val = $tag->{DECODER}($self,$val) if defined($tag->{DECODER});
+		$val = &{$tag->{DECODER}}($self,$val) if defined($tag->{DECODER});
 		$val = $tag->{$val} if exists $tag->{$val};
 		$tag = $tag->{__TAG__};
 	    }
@@ -567,7 +721,15 @@ sub as_string {
 
 sub as_float {
     my $self = shift;
-    $self->[0] / $self->[1];
+
+    # We check here because some stupid cameras (Samsung Digimax 200)
+    # use rationals with 0 denominator (found in thumbnail resolution spec).
+    if ($self->[1]) {
+	return $self->[0] / $self->[1];
+    }
+    else {
+ 	return $self->[0];
+    }
 }
 
 1;
